package ch.unige.solidify.test.helper;

import static org.springframework.restdocs.payload.PayloadDocumentation.fieldWithPath;

import java.util.List;
import java.util.stream.Collectors;

import org.apache.tomcat.util.buf.StringUtils;
import org.springframework.restdocs.payload.FieldDescriptor;
import org.springframework.restdocs.snippet.Attributes;

import ch.unige.solidify.test.model.TypeField;
import ch.unige.solidify.util.StringTool;

public class FieldDescriptorBuilder {

  public static final String CONSTRAINTS = "constraints";
  public static final String ENUM_VALUE_QUOTE = "`";
  public static final char ENUM_VALUE_SEPARATOR = ',';
  public static final String FORMAT_INTEGER_INT32 = "int32";
  public static final String FORMAT_INTEGER_INT64 = "int64";
  public static final String FORMAT_NUMBER_DOUBLE = "double";
  public static final String FORMAT_NUMBER_FLOAT = "float";
  public static final String FORMAT_PREFIX = "_Format_:";

  public static final String FORMAT_STRING_BINARY = "binary";
  public static final String FORMAT_STRING_BYTE = "byte";
  public static final String FORMAT_STRING_DATE = "date";
  public static final String FORMAT_STRING_DATE_TIME = "date-time";
  @SuppressWarnings("squid:S2068")
  public static final String FORMAT_STRING_PWD = "password";
  public static final String MANDATORY_VALUE = "mandatoryValue";
  public static final String NO_CONSTRAINT = "";
  public static final String READ_ONLY_VALUE = "readOnlyValue";
  public static final String TEST_STRING_VALUE = "test";
  public static final String VALUE_PREFIX = "_Values_:";
  protected static final String CONSTRAINT_CODE_ISO = "code ISO 3166-2";

  protected static final String CONSTRAINT_DATE_INTEGER = "3650 days ~ 10 years, 0 day ~ forever";
  protected static final String CONSTRAINT_VALID_DOI = "Valid DOI";
  protected static final String CONSTRAINT_VALID_ID = "Valid ID";
  protected static final String CONSTRAINT_VALID_URL = "Valid URL";
  protected static final String CONSTRAINT_VALID_JSON = "Valid JSON";
  protected static final String NO_TEST_VALUE = "NO_TEST_VALUE";

  protected static final String TEST_CODE_ISO = "CH";
  // Test value list
  protected static final String TEST_DATE_FORMAT_VALUE = "2018-07-10T15:36:50.323+0200";
  protected static final String TEST_DATE_SIMPLE_VALUE = "2018-07-10";
  protected static final String TEST_EMAIL_VALUE = "solidify@unige.ch";
  protected static final String TEST_NUMBER_VALUE = "11";
  protected static final String TEST_ORCID = "0000-1111-2222-3333";
  protected static final String TEST_SEARCH_QUERY = "*";
  protected static final String TEST_TRUE_VALUE = "true";
  protected static final String TEST_URL_VALUE = "https://www.unige.ch/";
  protected static final String TEST_JSON_OBJECT_VALUE = "{ \"test\" : \"value\" }";
  protected static final String TEST_JSON_ARRAY_VALUE = "[{ \"test\" : \"value\" }, { \"test2\" : \"value2\" }]";

  private AbstractResourceDoc resourceDoc;

  private String name;
  private String description;
  private boolean mandatory;
  private TypeField type = TypeField.STRING;
  private List<String> enumValues;
  private String customConstraints;
  private String testValue;
  private boolean readOnly = true;

  public FieldDescriptorBuilder(AbstractResourceDoc resourceDoc) {
    this.resourceDoc = resourceDoc;
  }

  public FieldDescriptorBuilder name(String name) {
    this.name = name;
    return this;
  }

  public FieldDescriptorBuilder description(String description) {
    this.description = description;
    return this;
  }

  public FieldDescriptorBuilder mandatory(boolean mandatory) {
    this.mandatory = mandatory;
    return this;
  }

  public FieldDescriptorBuilder type(TypeField type) {
    this.type = type;
    return this;
  }

  public FieldDescriptorBuilder enumValues(List<String> enumValues) {
    this.enumValues = enumValues;
    return this;
  }

  public FieldDescriptorBuilder customConstraints(String customConstraints) {
    this.customConstraints = customConstraints;
    return this;
  }

  public FieldDescriptorBuilder testValue(String testValue) {
    this.testValue = testValue;
    return this;
  }

  public FieldDescriptorBuilder readOnly(boolean readOnly) {
    this.readOnly = readOnly;
    return this;
  }

  public FieldDescriptor build() {
    final String constraints = this.determineConstraint(customConstraints, type, enumValues);
    testValue = this.determineTestValue(testValue, type, enumValues);
    final FieldDescriptor fieldDescriptor = fieldWithPath(name).description(description);
    this.setFieldDescriptorAttributes(fieldDescriptor, mandatory, testValue, constraints, readOnly);
    resourceDoc.addToOpenApiExtraDocumentation(name, enumValues, type);
    return fieldDescriptor;
  }

  private String determineConstraint(String constraints, TypeField type, List<String> enumValues) {
    if (constraints != null) {
      return constraints;
    }

    if (enumValues != null) {
      return this.listValues(enumValues);
    }

    switch (type) {
      case DATE:
        constraints = this.displayFormat(StringTool.DATE_FORMAT);
        break;
      case DATE_TIME:
        constraints = this.displayFormat(StringTool.DATE_TIME_FORMAT);
        break;
      default:
        constraints = NO_CONSTRAINT;
    }
    return constraints;
  }

  private String determineTestValue(String testValue, TypeField type, List<String> enumValues) {
    if (testValue != null) {
      if (testValue.equals(NO_TEST_VALUE)) {
        return null;
      }
      return testValue;
    }

    if (enumValues != null) {
      if (enumValues.isEmpty()) {
        return null;
      }
      return enumValues.get(0);
    }

    switch (type) {
      case INT32:
      case INT64:
      case FLOAT:
      case DOUBLE:
        testValue = TEST_NUMBER_VALUE;
        break;
      case STRING:
      case BYTE:
      case BINARY:
      case PASSWORD:
        testValue = TEST_STRING_VALUE;
        break;
      case BOOLEAN:
        testValue = TEST_TRUE_VALUE;
        break;
      case DATE:
        testValue = TEST_DATE_SIMPLE_VALUE;
        break;
      case DATE_TIME:
        testValue = TEST_DATE_FORMAT_VALUE;
        break;
      case ARRAY:
      case UNDEFINED:
        break;
      case URL:
        testValue = TEST_URL_VALUE;
        break;
    }
    return testValue;
  }

  private void setFieldDescriptorAttributes(FieldDescriptor fieldDescriptor, boolean mandatory,
          String testValue, String constraints, boolean readOnly) {
    fieldDescriptor.attributes(new Attributes.Attribute(CONSTRAINTS, constraints));
    // Add attribute only true because AsciiDoc test if the value is present
    if (mandatory) {
      fieldDescriptor.attributes(new Attributes.Attribute(MANDATORY_VALUE, String.valueOf(mandatory)));
    }
    fieldDescriptor.attributes(new Attributes.Attribute(READ_ONLY_VALUE, String.valueOf(readOnly)));

    this.setTestValue(fieldDescriptor, testValue);
  }

  private void setTestValue(FieldDescriptor fieldDescriptor, String testValue) {
    if (testValue == null) {
      return;
    }
    fieldDescriptor.attributes(new Attributes.Attribute(TEST_STRING_VALUE, testValue));
  }

  private String listValues(List<String> enumValues) {
    final List<String> enumValuesFormated = enumValues.stream()
            .map(val -> " " + ENUM_VALUE_QUOTE + val + ENUM_VALUE_QUOTE)
            .collect(Collectors.toList());
    return VALUE_PREFIX + StringUtils.join(enumValuesFormated, ENUM_VALUE_SEPARATOR);
  }

  protected String displayFormat(String format) {
    return FORMAT_PREFIX + " " + ENUM_VALUE_QUOTE + format + ENUM_VALUE_QUOTE;
  }

}
