package ch.unige.solidify.util;

import java.io.UnsupportedEncodingException;
import java.lang.reflect.Array;
import java.net.URI;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.nio.charset.StandardCharsets;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.time.OffsetDateTime;
import java.time.ZoneOffset;
import java.time.format.DateTimeFormatter;
import java.time.format.DateTimeParseException;
import java.util.ArrayList;
import java.util.Base64;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Objects;
import java.util.UUID;

import javax.validation.constraints.NotNull;
import javax.xml.bind.DatatypeConverter;

import org.json.JSONArray;
import org.json.JSONObject;
import org.json.JSONString;
import org.json.XML;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.fasterxml.jackson.databind.ObjectMapper;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyRuntimeException;

public class StringTool {
  public static final String DATE_FORMAT = "yyyy-MM-dd";
  public static final String DATE_TIME_FORMAT = "yyyy-MM-dd'T'HH:mm:ss.SSSZ";
  public static final String BYTES = "B";
  public static final String KILO_BYTES = "KB";
  public static final String MEGA_BYTES = "MB";
  public static final String GIGA_BYTES = "GB";
  public static final String TERA_BYTES = "TB";
  private static final Logger log = LoggerFactory.getLogger(StringTool.class);

  public static String capitalize(String str) {
    return StringTool.capitalize(str, true);
  }

  public static String capitalize(String str, boolean lowercase) {
    if (!StringTool.isNullOrEmpty(str)) {
      if (lowercase) {
        return str.substring(0, 1).toUpperCase() + str.substring(1).toLowerCase();
      } else {
        return str.substring(0, 1).toUpperCase() + str.substring(1);
      }
    } else {
      return str;
    }
  }

  /**
   * Check if the date is valid
   *
   * @param dateString the date as a string
   * @return true if valid false if not
   */
  public static boolean checkDate(String dateString) {
    if (dateString.matches("\\d{4}\\-\\d{2}\\-\\d{2}")) {
      dateString += "T00:00:00Z";
    }
    try {
      getDate(dateString);
      return true;
    } catch (final DateTimeParseException e) {
      return false;
    }
  }

  /**
   * Convert path with Windows File Separator (backslash) to path with Unix File Separator (slash)
   *
   * @param path the path to convert in unix path
   * @return path with slash as file separator
   */
  public static String convertInWindowsPath(String path) {
    if (path == null) {
      return null;
    }
    return path.replaceAll("\\\\", SolidifyConstants.URL_SEP);
  }

  public static String convertToQueryString(Object object) {
    return convertToQueryString(object, Collections.emptyList());
  }

  public static String convertToQueryString(Object object, @NotNull List<String> exceptionList) {
    Objects.requireNonNull(exceptionList, "The exception list must not be null");

    // Object --> map
    final ObjectMapper objectMapper = new ObjectMapper();
    @SuppressWarnings("unchecked")
    final Map<String, Object> map = objectMapper.convertValue(object, Map.class);

    final StringBuilder queryString = new StringBuilder();
    for (final Entry<String, Object> entry : map.entrySet()) {

      // Do not process if
      // empty object or HATEOAS links or Exception list

      if (entry.getValue() == null
              || entry.getKey().equals("links")
              || (!exceptionList.isEmpty() && exceptionList.contains(entry.getKey()))) {
        continue;
      }

      // Add & if needed
      if (queryString.length() > 0) {
        queryString.append("&");
      }
      // key=value
      queryString.append(entry.getKey());
      queryString.append("=");
      queryString.append(entry.getValue());
    }
    return queryString.toString();
  }

  public static String decode64(String encodedstring) {
    return new String(Base64.getDecoder().decode(encodedstring));
  }

  public static String decode64IfEncoded(String string) {
    if (string.startsWith(SolidifyConstants.BASE64_PREFIX)) {
      return StringTool.decode64(string.substring(SolidifyConstants.BASE64_PREFIX.length()));
    }
    return string;
  }

  public static String decodeUrl(String s) throws UnsupportedEncodingException {
    return URLDecoder.decode(s, StandardCharsets.UTF_8.toString());
  }

  public static String encode64(String string) {
    return Base64.getEncoder().withoutPadding().encodeToString(string.getBytes());
  }

  public static String encode64WithPadding(String string) {
    return Base64.getEncoder().encodeToString(string.getBytes());
  }

  public static String encodeUrl(String s) throws UnsupportedEncodingException {
    return URLEncoder.encode(s, StandardCharsets.UTF_8.toString());
  }

  public static void filterMap(Map<String, Object> metadata, String[] exceptions) {
    filterObject(metadata, exceptions);
  }

  public static String formatDate(OffsetDateTime date, DateTimeFormatter formatTemplate) {
    return date.toLocalDateTime().format(formatTemplate);
  }

  public static String formatSmartSize(long size) {
    String unit;
    double s;
    if (size < 1024l) {
      s = size;
      unit = BYTES;
    } else if (size < (1024l * 1024l)) {
      s = (double) size / 1024l;
      unit = KILO_BYTES;
    } else if (size < (1024l * 1024l * 1024l)) {
      s = (double) size / (1024l * 1024l);
      unit = MEGA_BYTES;
    } else if (size < (1024l * 1024l * 1024l * 1024l)) {
      s = (double) size / (1024l * 1024l * 1024l);
      unit = GIGA_BYTES;
    } else {
      s = (double) size / (1024l * 1024l * 1024l * 1024l);
      unit = TERA_BYTES;
    }
    return String.format(Locale.ROOT, "%.1f %s", s, unit);
  }

  public static long getSizeFromSmartSize(String smartSize) {
    double shortSize = Double.parseDouble(smartSize.substring(0, smartSize.indexOf(' ')));
    long size = 0;
    String unit = smartSize.substring(smartSize.indexOf(' ') + 1);
    switch (unit) {
      case BYTES:
        size = (long) shortSize;
        break;
      case KILO_BYTES:
        size = (long) (shortSize * 1024l);
        break;
      case MEGA_BYTES:
        size = (long) (shortSize * 1024l * 1024l);
        break;
      case GIGA_BYTES:
        size = (long) (shortSize * 1024l * 1024l * 1024l);
        break;
      case TERA_BYTES:
        size = (long) (shortSize * 1024l * 1024l * 1024l * 1024l);
        break;
      default:
        throw new SolidifyRuntimeException("Unknow unit " + unit);
    }
    return size;
  }

  public static String generateResId() {
    return UUID.randomUUID().toString();
  }

  /**
   * Get an OffsetDateTime instance from a String
   *
   * @param dateString the date as a string
   * @return the date as an OffsetDateTime
   * @throws DateTimeParseException
   */
  public static OffsetDateTime getDate(String dateString) {
    return OffsetDateTime.parse(dateString).toInstant().atOffset(ZoneOffset.UTC);
  }

  /*
   * Get an OffsetDateTime instance from a String and a DateTimeFormatter
   *
   * @param dateString the date as a string
   *
   * @param formatter the formatter
   *
   * @return the date as an OffsetDateTime
   *
   * @throws DateTimeParseException
   */
  public static OffsetDateTime getDate(String dateString, DateTimeFormatter formatter) {
    return OffsetDateTime.parse(dateString, formatter).toInstant().atOffset(ZoneOffset.UTC);
  }

  public static String getUrlPath(String url) {
    final URI uri = URI.create(url.trim());
    return uri.getPath();
  }

  public static String hash(String algo, String s) throws NoSuchAlgorithmException {
    final MessageDigest md = MessageDigest.getInstance(algo);
    md.update(s.getBytes());
    final byte[] digest = md.digest();
    return DatatypeConverter.printHexBinary(digest);
  }

  public static boolean isNullOrEmpty(String value) {
    return (value == null || value.isEmpty());
  }

  @SuppressWarnings("squid:S00117")
  public static String JSON2XML(String jsonString) {
    final JSONObject json = new JSONObject(jsonString);
    return XML.toString(json);
  }

  @SuppressWarnings("squid:S00117")
  public static String Map2JSON(Map<String, Object> metadata) {
    final JSONObject json = new JSONObject(metadata);
    return json.toString();
  }

  @SuppressWarnings("squid:S00117")
  public static String Map2XML(Map<String, Object> metadata) {
    return JSON2XML(Map2JSON(metadata));
  }

  public static String toCamelCase(String strTocamel) {
    if (isNullOrEmpty(strTocamel)) {
      return strTocamel;
    }
    return strTocamel.substring(0, 1).toUpperCase() + strTocamel.substring(1).toLowerCase();
  }

  public static String toHexaString(byte[] bytes) {
    // Convert it to hexadecimal format
    final StringBuilder sb = new StringBuilder();
    for (final byte aByte : bytes) {
      sb.append(Integer.toString((aByte & 0xff) + 0x100, 16).substring(1));
    }
    return sb.toString();
  }

  public static boolean toUpdate(String vOri, String vNew) {
    if (vNew == null || vNew.isEmpty()) {
      return false;
    }
    return (vOri == null || !vOri.equals(vNew));
  }

  public static String truncate(String str, int length) {
    if (StringTool.isNullOrEmpty(str) || length >= str.length() || length < 0) {
      return str;
    }
    return str.substring(0, length);
  }

  public static String substringBetween(String str, String startStr, String endStr) {
    if (str.contains(startStr)) {
      int startIndex = str.indexOf(startStr) + startStr.length();
      int endIndex = str.indexOf(endStr, startIndex);

      if (endIndex > -1) {
        str = str.substring(startIndex, endIndex);
      } else {
        str = str.substring(startIndex);
      }
    }
    return str;
  }

  public static int countSubstringOccurrences(String str, String subStr) {
    return (str.length() - str.replace(subStr, "").length()) / subStr.length();
  }

  public static String truncateWithEllipsis(String str, int length) {
    return StringTool.truncateWithEllipsis(str, length, "...");
  }

  public static String truncateWithEllipsis(String str, int length, String ellipsis) {
    if (StringTool.isNullOrEmpty(str) || length >= str.length() || length < 0
            || StringTool.isNullOrEmpty(ellipsis)) {
      return str;
    } else {
      final int lengthWithoutEllipsis = length - ellipsis.length();
      if (lengthWithoutEllipsis > 0) {
        return StringTool.truncate(str, lengthWithoutEllipsis) + ellipsis;
      } else {
        return StringTool.truncate(str, length);
      }
    }
  }

  @SuppressWarnings("squid:S00117")
  public static String XML2JSON(String xmlString) {
    final JSONObject json = XML.toJSONObject(xmlString);
    return json.toString();
  }

  private static void filterObject(Object obj, String[] exceptions) {
    if (obj == null) {
      return;
    }
    if (obj instanceof JSONObject || obj instanceof JSONArray || obj instanceof JSONString
            || obj instanceof Byte || obj instanceof Character || obj instanceof Short
            || obj instanceof Integer || obj instanceof Long || obj instanceof Boolean
            || obj instanceof Float || obj instanceof Double || obj instanceof String) {
      return;
    }

    if (obj instanceof Collection) {
      final Iterator<?> i = ((Collection<?>) obj).iterator();
      while (i.hasNext()) {
        filterObject(i.next(), exceptions);
      }
    }
    if (obj.getClass().isArray()) {
      final int length = Array.getLength(obj);
      for (int i = 0; i < length; i += 1) {
        filterObject(Array.get(obj, i), exceptions);
      }
    }
    if (obj instanceof Map) {
      final List<String> toDelete = new ArrayList<>();
      final Iterator<?> i = ((Map<?, ?>) obj).entrySet().iterator();
      while (i.hasNext()) {
        final Map.Entry<?, ?> e = (Map.Entry<?, ?>) i.next();
        final String k = ((String) e.getKey());
        // Do not index exception
        for (final String x7 : exceptions) {
          if (k.endsWith(x7)) {
            toDelete.add(k);
            break;
          }
        }
        // Remove empty object
        if (e.getValue() instanceof String) {
          final String v = ((String) e.getValue());
          if (StringTool.isNullOrEmpty(v)) {
            log.trace("To remove {}", k);
            toDelete.add(k);
          }
        } else {
          filterObject(e.getValue(), exceptions);
        }
      }
      for (final String k : toDelete) {
        log.trace("Removing {}", k);
        ((Map<?, ?>) obj).remove(k);
      }
    }
  }

  /**
   * This constructor should not be called
   */
  private StringTool() {
    throw new IllegalStateException(SolidifyConstants.TOOL_CLASS);
  }
}
