package ch.unige.solidify.test.util;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import org.hamcrest.core.IsEqual;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ErrorCollector;
import org.junit.runner.RunWith;
import org.springframework.test.context.junit4.SpringRunner;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyCheckingException;
import ch.unige.solidify.util.StringTool;

@RunWith(SpringRunner.class)
public class StringTest {
  @Rule
  public ErrorCollector collector = new ErrorCollector();

  @Test
  public void base64Test() {
    final String value = "Parameter with space and , and + and &";
    final String encodedValue = StringTool.encode64(value);
    final String encodedParam = SolidifyConstants.BASE64_PREFIX + encodedValue;

    assertTrue(StringTool.decode64(encodedValue).equals(value));
    assertTrue(StringTool.decode64IfEncoded(encodedParam).equals(value));
    assertTrue(StringTool.decode64IfEncoded(value).equals(value));
  }

  @Test
  public void dateTest() throws SolidifyCheckingException {
    this.collector.checkThat("Check YYYY-MM-DD", StringTool.checkDate("2018-01-24"),
            IsEqual.equalTo(true));
    this.collector.checkThat("Check YYYY-MM-DDTHH:MM:SSZ", StringTool.checkDate("2018-01-24T14:06:28Z"),
            IsEqual.equalTo(true));
    this.collector.checkThat("Check YYYY-MM-DDTHH:MM:SS.MMMZ",
            StringTool.checkDate("2018-01-24T14:06:28.000Z"), IsEqual.equalTo(true));
    this.collector.checkThat("Wrong date", StringTool.checkDate("2018-14-24T14:06:28.000Z"),
            IsEqual.equalTo(false));
  }

  @Test
  public void formatSmartSizeTest() {

    // @formatter:off
    final long b1       = 1;
    final long b10      = 10;
    final long b100     = 100;
    final long b1000    = 1000;

    final long kb1      = 1l    * 1024;
    final long kb10     = 10l   * 1024;
    final long kb100    = 100l  * 1024;
    final long kb1000   = 1000l * 1024;

    final long mega1    = 1l    * 1024 * 1024;
    final long mega10   = 10l   * 1024 * 1024;
    final long mega100  = 100l  * 1024 * 1024;
    final long mega1000 = 1000l * 1024 * 1024;

    final long gb1      = 1     * 1024 * 1024 * 1024;
    final long gb1_5    = 15    * 100  * 1024 * 1024;
    final long gb10     = 10l   * 1024 * 1024 * 1024;
    final long gb100    = 100l  * 1024 * 1024 * 1024;
    final long gb1000   = 1000l * 1024 * 1024 * 1024;

    final long tera1    = 1l    * 1024 * 1024 * 1024 * 1024;
    final long tera1_7  = 17l   * 100  * 1024 * 1024 * 1024;
    final long tera5    = 5l    * 1024 * 1024 * 1024 * 1024;
    final long tera10   = 10l   * 1024 * 1024 * 1024 * 1024;
    final long tera100  = 100l  * 1024 * 1024 * 1024 * 1024;
    final long tera1000 = 1000l * 1024 * 1024 * 1024 * 1024;
    // @formatter:on

    assertTrue(StringTool.formatSmartSize(b1).equals("1.0 B"));
    assertTrue(StringTool.formatSmartSize(b10).equals("10.0 B"));
    assertTrue(StringTool.formatSmartSize(b100).equals("100.0 B"));
    assertTrue(StringTool.formatSmartSize(b1000).equals("1000.0 B"));

    assertTrue(StringTool.formatSmartSize(kb1).equals("1.0 KB"));
    assertTrue(StringTool.formatSmartSize(kb10).equals("10.0 KB"));
    assertTrue(StringTool.formatSmartSize(kb100).equals("100.0 KB"));
    assertTrue(StringTool.formatSmartSize(kb1000).equals("1000.0 KB"));

    assertTrue(StringTool.formatSmartSize(mega1).equals("1.0 MB"));
    assertTrue(StringTool.formatSmartSize(mega10).equals("10.0 MB"));
    assertTrue(StringTool.formatSmartSize(mega100).equals("100.0 MB"));
    assertTrue(StringTool.formatSmartSize(mega1000).equals("1000.0 MB"));

    assertTrue(StringTool.formatSmartSize(gb1).equals("1.0 GB"));
    assertTrue(StringTool.formatSmartSize(gb1_5).equals("1.5 GB"));
    assertTrue(StringTool.formatSmartSize(gb10).equals("10.0 GB"));
    assertTrue(StringTool.formatSmartSize(gb100).equals("100.0 GB"));
    assertTrue(StringTool.formatSmartSize(gb1000).equals("1000.0 GB"));

    assertTrue(StringTool.formatSmartSize(tera1).equals("1.0 TB"));
    assertTrue(StringTool.formatSmartSize(tera1_7).equals("1.7 TB"));
    assertTrue(StringTool.formatSmartSize(tera5).equals("5.0 TB"));
    assertTrue(StringTool.formatSmartSize(tera10).equals("10.0 TB"));
    assertTrue(StringTool.formatSmartSize(tera100).equals("100.0 TB"));
    assertTrue(StringTool.formatSmartSize(tera1000).equals("1000.0 TB"));
  }

  @Test
  public void getSizeFromSmartSizeTest() {

    // @formatter:off
    final long b1       = 1;
    final long b10      = 10;
    final long b100     = 100;
    final long b1000    = 1000;

    final long kb1      = 1l    * 1024;
    final long kb10     = 10l   * 1024;
    final long kb100    = 100l  * 1024;
    final long kb1000   = 1000l * 1024;

    final long mega1    = 1l    * 1024 * 1024;
    final long mega10   = 10l   * 1024 * 1024;
    final long mega100  = 100l  * 1024 * 1024;
    final long mega1000 = 1000l * 1024 * 1024;
    // @formatter:on

    assertEquals(b1, StringTool.getSizeFromSmartSize(StringTool.formatSmartSize(b1)));
    assertEquals(b10, StringTool.getSizeFromSmartSize(StringTool.formatSmartSize(b10)));
    assertEquals(b100, StringTool.getSizeFromSmartSize(StringTool.formatSmartSize(b100)));
    assertEquals(b1000, StringTool.getSizeFromSmartSize(StringTool.formatSmartSize(b1000)));

    assertEquals(kb1, StringTool.getSizeFromSmartSize(StringTool.formatSmartSize(kb1)));
    assertEquals(kb10, StringTool.getSizeFromSmartSize(StringTool.formatSmartSize(kb10)));
    assertEquals(kb100, StringTool.getSizeFromSmartSize(StringTool.formatSmartSize(kb100)));
    assertEquals(kb1000, StringTool.getSizeFromSmartSize(StringTool.formatSmartSize(kb1000)));

    assertEquals(mega1, StringTool.getSizeFromSmartSize(StringTool.formatSmartSize(mega1)));
    assertEquals(mega10, StringTool.getSizeFromSmartSize(StringTool.formatSmartSize(mega10)));
    assertEquals(mega100, StringTool.getSizeFromSmartSize(StringTool.formatSmartSize(mega100)));
    assertEquals(mega1000, StringTool.getSizeFromSmartSize(StringTool.formatSmartSize(mega1000)));
  }

  @Test
  public void truncateTest() {

    // truncate if too long
    final String longStr = "long string that we need to truncate";
    String truncatedStr = StringTool.truncate(longStr, 20);
    assertTrue(truncatedStr.length() == 20);
    assertTrue(truncatedStr.equals(longStr.substring(0, 20)));

    // do not truncate if shorter
    truncatedStr = StringTool.truncate(longStr, 100);
    assertTrue(truncatedStr.equals(longStr));

    // do not do anything if length is wrong
    truncatedStr = StringTool.truncate(longStr, -10);
    assertTrue(truncatedStr.equals(longStr));
  }

  @Test
  public void truncateWithCustomEllipsisTest() {

    final String longStr = "long string that we need to truncate";
    final String shortStr = "abcdef";
    final String ellipsis = "****";

    // truncate if too long
    String truncatedStr = StringTool.truncateWithEllipsis(longStr, 20, ellipsis);
    assertTrue(truncatedStr.length() == 20);
    assertFalse(truncatedStr.equals(longStr.substring(0, 20)));
    assertTrue(truncatedStr.equals(longStr.substring(0, 20 - ellipsis.length()) + ellipsis));

    // do not truncate if shorter than allowed
    truncatedStr = StringTool.truncateWithEllipsis(longStr, 100, ellipsis);
    assertTrue(truncatedStr.equals(longStr));

    // do not do anything if length is wrong
    truncatedStr = StringTool.truncateWithEllipsis(longStr, -10, ellipsis);
    assertTrue(truncatedStr.equals(longStr));

    // truncate if too long and some original string is left
    truncatedStr = StringTool.truncateWithEllipsis(shortStr, 5, ellipsis);
    assertTrue(truncatedStr.equals("a" + ellipsis));

    // do not use ellipsis if no original string would remain
    truncatedStr = StringTool.truncateWithEllipsis(shortStr, 3, ellipsis);
    assertTrue(truncatedStr.equals("abc"));

    // do not use ellipsis if no original string would remain
    truncatedStr = StringTool.truncateWithEllipsis(shortStr, 2, ellipsis);
    assertTrue(truncatedStr.equals("ab"));
  }

  @Test
  public void truncateWithEllipsisTest() {

    final String longStr = "long string that we need to truncate";
    final String shortStr = "abcde";

    // truncate if too long
    String truncatedStr = StringTool.truncateWithEllipsis(longStr, 20);
    assertTrue(truncatedStr.length() == 20);
    assertFalse(truncatedStr.equals(longStr.substring(0, 20)));
    assertTrue(truncatedStr.equals(longStr.substring(0, 17) + "..."));

    // do not truncate if shorter than allowed
    truncatedStr = StringTool.truncateWithEllipsis(longStr, 100);
    assertTrue(truncatedStr.equals(longStr));

    // do not do anything if length is wrong
    truncatedStr = StringTool.truncateWithEllipsis(longStr, -10);
    assertTrue(truncatedStr.equals(longStr));

    // truncate if too long and some original string is left
    truncatedStr = StringTool.truncateWithEllipsis(shortStr, 4);
    assertTrue(truncatedStr.equals("a..."));

    // do not use ellipsis if no original string would remain
    truncatedStr = StringTool.truncateWithEllipsis(shortStr, 3);
    assertTrue(truncatedStr.equals("abc"));

    // do not use ellipsis if no original string would remain
    truncatedStr = StringTool.truncateWithEllipsis(shortStr, 2);
    assertTrue(truncatedStr.equals("ab"));
  }

  @Test
  public void substringBetweenTest() {
    final String longStr = "a string from which [start]we want to extract[end] a substring";
    final String startStr = "[start]";
    final String endStr = "[end]";

    String subStr = StringTool.substringBetween(longStr, startStr, endStr);
    assertTrue(subStr.equals("we want to extract"));

    subStr = StringTool.substringBetween(longStr, startStr, "[xxx]");
    assertTrue(subStr.equals("we want to extract[end] a substring"));

    subStr = StringTool.substringBetween(longStr, endStr, startStr);
    assertTrue(subStr.equals(" a substring"));

    subStr = StringTool.substringBetween(longStr, "[xxx]", endStr);
    assertTrue(subStr.equals(longStr));

    subStr = StringTool.substringBetween(longStr, "[xxx]", "[yyy]");
    assertTrue(subStr.equals(longStr));
  }

  @Test
  public void countSubstringOccurrences() {
    String sub = "zzz";
    assertEquals(0, StringTool.countSubstringOccurrences("Lorem ipsum dolor sit amet,  elit, sed do eiusmod", sub));
    assertEquals(1, StringTool.countSubstringOccurrences("Lorem sit " + sub + " amet, consectetur adipisicing elit,", sub));
    assertEquals(2, StringTool.countSubstringOccurrences("dolor sit " + sub + " amet, adipisi" + sub + "cing elit", sub));
    assertEquals(1, StringTool.countSubstringOccurrences(sub + "Lorem ipsum dolor sit amet, sed do eiusmod", sub));
    assertEquals(1, StringTool.countSubstringOccurrences("Lorem ipsum dolor sit amet,, sed do eiusmod" + sub, sub));
    assertEquals(3, StringTool.countSubstringOccurrences(sub + "Lorem " + sub + "sit amet" + "elit, " + sub + "sed do", sub));
    assertEquals(3, StringTool.countSubstringOccurrences("Lorem " + sub + " ipsum dolor " + sub + "sit amet, eiusmod" + sub, sub));
    assertEquals(4, StringTool.countSubstringOccurrences(sub + "Lorem " + sub + " ipsum dolor " + sub + "sit eiusmod" + sub, sub));
  }
}
